from __future__ import annotations

# class VerificationDataSource:
#     """
#     Groups all contracts for a specific data_source. Used during contract verification execution to group all
#     contracts per data_source and ensure the data_source is open during verification of the contract for this data_source.
#     """
#
#     def __init__(self) -> None:
#         self.data_source: DataSource | None = None
#         self.contracts: list[Contract] = []
#
#     def requires_with_block(self) -> bool:
#         return True
#
#     def add_contract(self, contract: Contract) -> None:
#         self.contracts.append(contract)
#
#
# class FileVerificationDataSource(VerificationDataSource):
#     def __init__(self, data_source_yaml_file: YamlFile):
#         super().__init__()
#         self.data_source = DataSource.from_yaml_file(data_source_yaml_file)
#
#
# class SparkConfiguration:
#     def __init__(self, spark_session: object, data_source_yaml_dict: dict, logs: Logs):
#         self.spark_session: object = spark_session
#         self.data_source_yaml_dict: dict = data_source_yaml_dict
#         self.logs: Logs = logs
#
#
# class SparkVerificationDataSource(VerificationDataSource):
#     def __init__(self, spark_configuration: SparkConfiguration):
#         super().__init__()
#         self.data_source = DataSource.from_spark_session(
#             spark_session=spark_configuration.spark_session,
#             data_source_yaml_dict=spark_configuration.data_source_yaml_dict,
#             logs=spark_configuration.logs
#         )



from __future__ import annotations

import importlib
from abc import ABC, abstractmethod

from soda.contracts.contract import ContractResult
from soda.contracts.impl.logs import Logs
from soda.contracts.impl.yaml_helper import YamlFile


class Plugin(ABC):

    @classmethod
    def create(cls, plugin_name: str, plugin_yaml_files: list[YamlFile], logs: Logs) -> Plugin | None:
        # dynamically load plugin based on name and feed plugin_yaml_files in the constructor
        try:
            module = importlib.import_module(f"soda.{plugin_name}.{plugin_name}_plugin")
            plugin_class_name = cls.get_plugin_class_name(plugin_name)
            class_ = getattr(module, plugin_class_name)
            return class_(logs, plugin_name, plugin_yaml_files)
        except ModuleNotFoundError as e:
            logs.error(f"Could not instantiate plugin {plugin_name}: {e}")
            return None

    def __init__(self, logs: Logs, plugin_name: str, plugin_yaml_files: list[YamlFile]):
        self.logs: Logs = logs
        self.plugin_name = plugin_name
        self.plugin_yaml_files: list[YamlFile] = plugin_yaml_files

    @classmethod
    def get_plugin_class_name(cls, plugin_name: str) -> str:
        # if "integration" == plugin_name:
        #     return "IntegrationPlugin"
        return f"{plugin_name[0:1].upper()}{plugin_name[1:]}Plugin"

    @abstractmethod
    def process_contract_results(self, contract_result: ContractResult) -> None:
        pass



from __future__ import annotations

import traceback
from dataclasses import dataclass
from enum import Enum
from textwrap import indent
from typing import List


@dataclass
class Location:

    file_path: str | None
    line: int | None
    column: int | None

    def __str__(self):
        parts = [
            f"line={self.line}" if self.line is not None else None,
            f"column={self.column}" if self.column is not None else None,
            f"file={self.file_path}" if self.file_path is not None else None,
        ]
        parts = [p for p in parts if p is not None]
        return ",".join(parts)

    def __hash__(self) -> int:
        return hash((self.line, self.column))


class LogLevel(Enum):
    ERROR = "error"
    WARNING = "warning"
    INFO = "info"
    DEBUG = "debug"


class Log:

    def __init__(
        self,
        level: LogLevel,
        message: str,
        location: Location | None = None,
        exception: BaseException | None = None,
        docs: str | None = None,
    ):
        self.level: LogLevel = level
        self.message: str = message
        self.location: Location | None = location
        self.exception: Exception | None = exception
        self.docs: str | None = docs

    def __str__(self):
        return self.to_string(include_stacktraces=True)

    def to_string(self, include_stacktraces: bool = False) -> str:
        location_str = f" | {self.location}" if self.location else ""
        doc_str = f" | https://go.soda.io/{self.docs}" if self.docs else ""
        exception_str = ""
        if self.exception:
            stacktrace_str = ""
            if include_stacktraces:
                stacktrace_str = "".join(traceback.format_tb(self.exception.__traceback__))
                stacktrace_str = stacktrace_str.strip()
            exception_str = f" | {self.exception}\n{indent(text=stacktrace_str, prefix='    ')}"
        return f"{self.level.value.ljust(7)}| {self.message}{location_str}{doc_str}{exception_str}"

    @classmethod
    def error(cls, message: str, location: Location | None = None, exception: BaseException | None = None) -> Log:
        return Log(level=LogLevel.ERROR, message=message, location=location, exception=exception)


class Logs:

    # See also adr/03_exceptions_vs_error_logs.md

    def __init__(self, logs: Logs | None = None):
        self.logs: List[Log] = []
        if logs is not None:
            self.logs = logs.logs.copy()

    def __str__(self) -> str:
        return "\n".join([str(log) for log in self.logs])

    def has_errors(self) -> bool:
        return any(log.level == LogLevel.ERROR for log in self.logs)

    def get_errors_str(self) -> str:
        errors_lines: List[str] = [str(log) for log in self.logs if log.level == LogLevel.ERROR]
        error_text = "\n".join(errors_lines)
        error_word = "Error: " if len(self.logs) == 1 else "Errors:\n"
        return f"{error_word}{error_text}"

    def get_errors(self) -> List[Log]:
        return [log for log in self.logs if log.level == LogLevel.ERROR]

    def error(self, message: str, location: Location | None = None, exception: BaseException | None = None) -> None:
        self._log(Log(LogLevel.ERROR, message, location, exception))

    def info(self, message: str, location: Location | None = None, exception: BaseException | None = None) -> None:
        self._log(Log(LogLevel.INFO, message, location, exception))

    def _log(self, log: Log) -> None:
        self.logs.append(log)



from __future__ import annotations

from hashlib import blake2b
from numbers import Number


class ConsistentHashBuilder:
    def __init__(self, hash_string_length: int = 8):
        if hash_string_length % 2 != 0:
            raise AssertionError(f"hash_string_length must be divisible by 2: {hash_string_length} is not")
        self.hash_string_length = hash_string_length
        self.blake2b = None

    def __get_blake2b(self) -> blake2b:
        # Lazy initialization of blake2b in order to return None in the self.get_hash(self) in case nothing was added
        if self.blake2b is None:
            self.blake2b = blake2b(digest_size=int(self.hash_string_length / 2))
        return self.blake2b

    def add(self, value: object | None) -> ConsistentHashBuilder:
        if value is not None:
            if isinstance(value, str):
                self.__get_blake2b().update(value.encode("utf-8"))
            elif isinstance(value, dict):
                for key, value in value.items():
                    self.add_property(key, value)
            elif isinstance(value, list):
                for e in value:
                    self.add(e)
            elif isinstance(value, Number) or isinstance(value, bool):
                self.__get_blake2b().update(str(value).encode("utf-8"))
            else:
                raise AssertionError(f"Unsupported hash value type {value} ({type(value).__name__})")
        return self

    def add_property(self, key: str, value: object | None) -> ConsistentHashBuilder:
        if value is not None:
            self.add(key)
            self.add(value)
        return self

    def get_hash(self) -> str:
        return self.blake2b.hexdigest() if self.blake2b else None



from __future__ import annotations

import os
from numbers import Number

from ruamel.yaml import YAML, CommentedMap, CommentedSeq, round_trip_dump
from ruamel.yaml.error import MarkedYAMLError

from soda.contracts.impl.logs import Location, Logs
from soda.contracts.impl.variable_resolver import VariableResolver


class YamlFile:
    """
    Allows yaml configurations to be specified as a file_path, yaml_str or a yaml_dict.
    If yaml configurations are specified as text (file or str), then parsing will resolve variables.
    Then YamlFile provides access to resolved and parsed dict.
    Usage:
    variables: dict = {...}
    yaml_file: YamlFile = YamlFile(logs=logs, ...use 1 of the other arguments to specify the yaml file content...)
    yaml_file.parse(variables)
    yaml_dict: dict = yaml_file.get_dict()
    """

    def __init__(
        self,
        logs: Logs,
        yaml_file_path: str | None = None,
        yaml_str: str | None = None,
        yaml_dict: dict | None = None,
    ):
        self.file_path: str | None = yaml_file_path
        self.source_str: str | None = yaml_str
        self.resolved_str: str | None = None
        self.dict: dict | None = yaml_dict
        self.is_parsed: bool = self.file_path is None and self.source_str is None and isinstance(self.dict, dict)
        self.logs: Logs = logs

    def get_dict(self) -> dict:
        assert self.is_parsed, "Usage of YamlFile requires that self.parse(...) " "is invoked before the dict is used."
        return self.dict

    def is_ok(self):
        """
        True if the yaml dict is available and the parsing didn't have errors
        """
        return isinstance(self.dict, dict)

    def parse(self, variables: dict) -> bool:
        """
        returns is_ok() indicating that parsing was successful and the dict is available
        """
        self.is_parsed = True

        if self.file_path is None and self.source_str is None and self.dict is None:
            self.logs.error("File not configured")

        if isinstance(self.file_path, str) and self.source_str is None:
            self.source_str = self.__read_file_as_str(file_path=self.file_path, logs=self.logs)

        self.resolved_str = self.__resolve_variables(source_str=self.source_str, variables=variables, logs=self.logs)

        if isinstance(self.resolved_str, str) and self.dict is None:
            self.dict = self.__parse_yaml_str(yaml_str=self.resolved_str, logs=self.logs)

        # It is assumed that if this parse is not ok, that an error has been logged
        return self.is_ok()

    @classmethod
    def __read_file_as_str(cls, file_path: str | None, logs: Logs) -> str | None:
        try:
            with open(file_path) as f:
                return f.read()
        except OSError as e:
            if not os.path.exists(file_path):
                logs.error(f"File '{file_path}' does not exist")
            elif not os.path.isdir(file_path):
                logs.error(f"File path '{file_path}' is a directory")
            else:
                logs.error(f"File '{file_path}' can't be read: {e}")

    @classmethod
    def __resolve_variables(cls, source_str: str | None, variables: dict[str, str] | None, logs: Logs) -> str | None:
        if isinstance(source_str, str):
            # Resolve all the ${VARIABLES} in the contract based on either the provided
            # variables or system variables (os.environ)
            variable_resolver = VariableResolver(logs=logs, variables=variables)
            return variable_resolver.resolve(source_str)
        else:
            return source_str

    def __parse_yaml_str(self, yaml_str: str | None, logs: Logs) -> dict | None:
        try:
            ruamel_yaml: YAML = YAML()
            ruamel_yaml.preserve_quotes = True
            return ruamel_yaml.load(yaml_str)
        except MarkedYAMLError as e:
            mark = e.context_mark if e.context_mark else e.problem_mark
            line = mark.line + 1
            col = mark.column + 1
            location = Location(file_path=self.get_file_description(), line=line, column=col)
            logs.error(f"YAML syntax error: {e}", location)

    def get_file_description(self) -> str:
        if self.file_path:
            return self.file_path
        if self.source_str:
            return "provided YAML str"
        if self.dict:
            return "provided dict"
        return "no yaml source provided"

    def exists(self):
        if self.file_path:
            return os.path.isfile(self.file_path)
        return isinstance(self.source_str, str) or isinstance(self.dict, dict)


class YamlHelper:

    def __init__(self, logs: Logs, yaml_file: YamlFile | None = None) -> None:
        self.logs: Logs = logs
        self.yaml_file: YamlFile | None = yaml_file

    def write_to_yaml_str(self, yaml_object: object) -> str:
        try:
            return round_trip_dump(yaml_object)
        except Exception as e:
            self.logs.error(f"Couldn't write SodaCL YAML object: {e}", exception=e)

    def create_location_from_yaml_dict_key(self, d: dict, key) -> Location | None:
        if isinstance(d, CommentedMap):
            if key in d:
                ruamel_location = d.lc.value(key)
                line: int = ruamel_location[0]
                column: int = ruamel_location[1]
                return Location(file_path=self.yaml_file.get_file_description(), line=line, column=column)
            else:
                return self.create_location_from_yaml_value(d)
        return None

    def create_location_from_yaml_value(self, d: object) -> Location | None:
        if isinstance(d, CommentedMap) or isinstance(d, CommentedSeq):
            return Location(file_path=self.yaml_file.get_file_description(), line=d.lc.line, column=d.lc.col)
        return None

    def read_dict(self, d: dict, key: str) -> dict | None:
        """
        An error is generated if the value is missing or not a YAML object.
        :return: a dict if the value for the key is a YAML object, otherwise None.
        """
        return self.read_value(d=d, key=key, expected_type=dict, required=True, default_value=None)

    def read_dict_opt(self, d: dict, key: str) -> dict | None:
        """
        An error is generated if the value is present and not a YAML object.
        :return: a dict if the value for the key is a YAML object, otherwise None.
        """
        return self.read_value(d=d, key=key, expected_type=dict, required=False, default_value=None)

    def read_list(self, d: dict, key: str) -> list | None:
        """
        An error is generated if the value is missing or not a YAML list.
        :return: a list if the value for the key is a YAML list, otherwise None.
        """
        return self.read_value(d=d, key=key, expected_type=list, required=True, default_value=None)

    def read_list_opt(self, d: dict, key: str) -> list | None:
        """
        An error is generated if the value is present and not a YAML list.
        :return: a list if the value for the key is a YAML list, otherwise None.
        """
        return self.read_value(d=d, key=key, expected_type=list, required=False, default_value=None)

    def read_list_of_dicts(self, d: dict, key: str) -> list[dict] | None:
        list_value: list = self.read_list(d, key)
        if isinstance(list_value, list):
            if all(isinstance(e, dict) for e in list_value):
                return list_value
            else:
                location: Location = self.create_location_from_yaml_dict_key(d, key)
                self.logs.error(f"Not all elements in list '{key}' are objects", location=location)

    def read_list_of_strings(self, d: dict, key: str) -> list[str] | None:
        list_value = self.read_value(d=d, key=key, expected_type=list, required=True, default_value=None)
        if isinstance(list_value, list):
            if all(isinstance(e, str) for e in list_value):
                return list_value
            else:
                location: Location | None = self.create_location_from_yaml_dict_key(d, key)
                self.logs.error(message=f"Not all elements in list '{key}' are strings", location=location)

    def read_string(self, d: dict, key: str) -> str | None:
        """
        An error is generated if the value is missing or not a string.
        :return: a str if the value for the key is a YAML string, otherwise None.
        """
        return self.read_value(d=d, key=key, expected_type=str, required=True, default_value=None)

    def read_string_opt(self, d: dict, key: str, default_value: str | None = None) -> str | None:
        """
        An error is generated if the value is present and not a string.
        :return: a str if the value for the key is a string, otherwise None.
        """
        return self.read_value(d=d, key=key, expected_type=str, required=False, default_value=default_value)

    def read_range(self, d: dict, key: str):  # returns Range | None
        range_yaml: list | None = self.read_list_opt(d, key)
        if isinstance(range_yaml, list):
            if all(isinstance(range_value, Number) for range_value in range_yaml) and len(range_yaml) == 2:
                from soda.contracts.check import Range

                return Range(lower_bound=range_yaml[0], upper_bound=range_yaml[1])
            else:
                location: Location = self.create_location_from_yaml_value(range_yaml)
                self.logs.error("range expects a list of 2 numbers", location=location)

    def read_bool(self, d: dict, key: str) -> bool | None:
        """
        An appropriate error log is generated if the value is not a bool or if the key is missing
        :return: a bool if the value for the key is a YAML boolean, otherwise None.
        """
        return self.read_value(d=d, key=key, expected_type=bool, required=True, default_value=None)

    def read_bool_opt(self, d: dict, key: str, default_value: bool | None = None) -> bool | None:
        """
        An appropriate error log is generated if the value is not a bool.
        :return: a bool if the value for the key is a YAML boolean, otherwise None.
        """
        return self.read_value(d=d, key=key, expected_type=bool, required=False, default_value=default_value)

    def read_number(self, d: dict, key: str) -> Number | None:
        """
        An appropriate error log is generated if the value is not a number or if the key is missing
        :return: a bool if the value for the key is a YAML number, otherwise None.
        """
        return self.read_value(d=d, key=key, expected_type=Number, required=True, default_value=None)

    def read_number_opt(self, d: dict, key: str, default_value: Number | None = None) -> Number | None:
        """
        An appropriate error log is generated if the value is not a number.
        :return: a Number if the value for the key is a YAML number, otherwise None.
        """
        return self.read_value(d=d, key=key, expected_type=Number, required=False, default_value=default_value)

    def read_value(
        self,
        d: dict,
        key: str,
        expected_type: type = None,
        required: bool = False,
        default_value=None,
    ) -> object | None:
        if key not in d:
            if required:
                location = self.create_location_from_yaml_dict_key(d, key)
                self.logs.error(message=f"'{key}' is required", location=location)
            return default_value
        value = d.get(key)
        if expected_type is not None and not isinstance(value, expected_type):
            location = self.create_location_from_yaml_dict_key(d, key)
            self.logs.error(
                message=f"'{key}' expected a {expected_type.__name__}, but was {type(value).__name__}",
                location=location,
            )
        return value


class QuotingSerializer:

    @classmethod
    def quote(cls, name: str) -> str:
        return (
            f'"{name}"'
            # Depends on ruamel class names DoubleQuotedScalarString and SingleQuotedScalarString
            if isinstance(name, str) and "Quoted" in type(name).__name__
            else name
        )



from __future__ import annotations

import logging
from dataclasses import dataclass
from numbers import Number
from textwrap import indent
from typing import List

from soda.scan import Scan

from soda.contracts.check import (
    AbstractCheck,
    Check,
    CheckArgs,
    CheckFactory,
    CheckOutcome,
    CheckResult,
    DuplicateCheckFactory,
    FreshnessCheckFactory,
    InvalidCheckFactory,
    MissingCheckFactory,
    MissingConfigurations,
    MultiColumnDuplicateCheckFactory,
    RowCountCheckFactory,
    SchemaCheck,
    SqlFunctionCheckFactory,
    Threshold,
    UserDefinedMetricExpressionCheckFactory,
    UserDefinedMetricQueryCheckFactory,
    ValidConfigurations,
    ValidValuesReferenceData,
)
from soda.contracts.impl.contract_data_source import ContractDataSource
from soda.contracts.impl.json_schema_verifier import JsonSchemaVerifier
from soda.contracts.impl.logs import Location, Logs
from soda.contracts.impl.yaml_helper import YamlFile, YamlHelper

logger = logging.getLogger(__name__)


class Contract:

    def __init__(
        self,
        contract_file: YamlFile,
        logs: Logs,
    ):
        self.contract_file: YamlFile = contract_file
        self.logs: Logs = logs

        self.data_source_name: str | None = None
        self.database_name: str | None = None
        self.schema_name: str | None = None
        self.dataset_name: str | None = None

        # TODO explain filter_expression_sql, default filter and named filters
        # filter name must part of the identity of the metrics
        #   - no filter part if no filter is specified
        #   - "default" is the filter name if there is only a default specified with "filter_expression_sql"
        #   - {filter_name} if a filter is activated from a named map of filters
        self.filter: str | None = None

        self.filter_sql: str | None = None
        self.checks: list[Check] = []

        self.missing_value_configs_by_column: dict[str, MissingConfigurations] = {}
        self.valid_value_configs_by_column: dict[str, ValidConfigurations] = {}

        self.__parse()

    def __parse(self) -> None:
        """
        Dry run: parse but not verify the contract to get the errors in the logs.
        """
        try:
            yaml_helper = YamlHelper(yaml_file=self.contract_file, logs=self.logs)

            if not self.contract_file.is_ok():
                return

            # Verify the contract schema on the ruamel instance object
            json_schema_verifier: JsonSchemaVerifier = JsonSchemaVerifier(self.logs)
            json_schema_verifier.verify(self.contract_file.get_dict())

            contract_yaml_dict = self.contract_file.get_dict()

            self.data_source_name: str | None = yaml_helper.read_string_opt(contract_yaml_dict, "data_source")
            self.database_name: str | None = yaml_helper.read_string_opt(contract_yaml_dict, "database")
            self.schema_name: str | None = yaml_helper.read_string_opt(contract_yaml_dict, "schema")
            self.dataset_name: str | None = yaml_helper.read_string(contract_yaml_dict, "dataset")
            self.filter_sql: str | None = yaml_helper.read_string_opt(contract_yaml_dict, "filter_sql")
            self.filter: str | None = "default" if self.filter_sql else None

            self.checks.append(
                SchemaCheck(
                    contract_file=self.contract_file,
                    data_source_name=self.data_source_name,
                    database_name=self.database_name,
                    schema_name=self.schema_name,
                    dataset_name=self.dataset_name,
                    yaml_contract=contract_yaml_dict,
                    logs=self.logs,
                )
            )

            column_yamls: list | None = yaml_helper.read_list(contract_yaml_dict, "columns")
            if column_yamls:
                for column_yaml in column_yamls:
                    column: str | None = yaml_helper.read_string(column_yaml, "name")
                    check_yamls: list | None = yaml_helper.read_list_opt(column_yaml, "checks")
                    if column and check_yamls:
                        for check_yaml in check_yamls:
                            self.__parse_column_check(check_yaml, column, yaml_helper)

            check_yamls: list | None = yaml_helper.read_list_opt(contract_yaml_dict, "checks")
            if check_yamls:
                for check_yaml in check_yamls:
                    self.__parse_dataset_check(check_yaml, yaml_helper)

            checks_by_identity: dict[str, Check] = {}
            for check in self.checks:
                if check.identity in checks_by_identity:
                    other_check: Check = checks_by_identity[check.identity]
                    if other_check:
                        location_info: str = ""
                        if isinstance(check, AbstractCheck) and isinstance(other_check, AbstractCheck):
                            location_info = f": {other_check.location} and {check.location}"
                        self.logs.error(f"Duplicate check identity '{check.identity}'{location_info}")
                else:
                    checks_by_identity[check.identity] = check

        except Exception as e:
            self.logs.error(message=f"Could not verify contract: {e}", exception=e)

    def __parse_dataset_check(self, check_yaml: dict, yaml_helper: YamlHelper):
        check_type: str | None = yaml_helper.read_string(check_yaml, "type")
        check_name = yaml_helper.read_string_opt(check_yaml, "name")
        check_name_was = yaml_helper.read_string_opt(check_yaml, "name_was")
        check_filter_sql = yaml_helper.read_string_opt(check_yaml, "filter_sql")
        threshold: Threshold = self.__parse_numeric_threshold(check_yaml=check_yaml)
        location: Location = yaml_helper.create_location_from_yaml_value(check_yaml)
        check_args: CheckArgs = CheckArgs(
            logs=self.logs,
            contract_file=self.contract_file,
            data_source_name=self.data_source_name,
            database_name=self.database_name,
            schema_name=self.schema_name,
            dataset_name=self.dataset_name,
            filter=self.filter,
            check_type=check_type,
            check_yaml=check_yaml,
            check_name=check_name,
            check_name_was=check_name_was,
            check_filter_sql=check_filter_sql,
            threshold=threshold,
            location=location,
            yaml_helper=yaml_helper,
        )
        dataset_check_factory_classes: list[CheckFactory] = [
            UserDefinedMetricExpressionCheckFactory(),
            UserDefinedMetricQueryCheckFactory(),
            RowCountCheckFactory(),
            MultiColumnDuplicateCheckFactory(),
        ]
        check: Check = self.__create_check(check_args, dataset_check_factory_classes)
        if check:
            self.checks.append(check)
        else:
            self.logs.error(message=f"Invalid dataset check {check_args.check_type}", location=check_args.location)

    def __parse_column_check(self, check_yaml: dict, column: str, yaml_helper: YamlHelper) -> None:
        check_type: str | None = yaml_helper.read_string(check_yaml, "type")
        check_name = yaml_helper.read_string_opt(check_yaml, "name")
        check_name_was = yaml_helper.read_string_opt(check_yaml, "name_was")
        check_filter_sql = yaml_helper.read_string_opt(check_yaml, "filter_sql")
        missing_configurations: MissingConfigurations | None = self.__parse_missing_configurations(
            check_yaml=check_yaml, column=column
        )
        valid_configurations: ValidConfigurations | None = self.__parse_valid_configurations(
            check_yaml=check_yaml, column=column
        )
        threshold: Threshold = self.__parse_numeric_threshold(check_yaml=check_yaml)
        location: Location = yaml_helper.create_location_from_yaml_value(check_yaml)
        check_args: CheckArgs = CheckArgs(
            logs=self.logs,
            contract_file=self.contract_file,
            data_source_name=self.data_source_name,
            database_name=self.database_name,
            schema_name=self.schema_name,
            dataset_name=self.dataset_name,
            filter=self.filter,
            check_type=check_type,
            check_yaml=check_yaml,
            check_name=check_name,
            check_name_was=check_name_was,
            check_filter_sql=check_filter_sql,
            threshold=threshold,
            location=location,
            yaml_helper=yaml_helper,
            column=column,
            missing_configurations=missing_configurations,
            valid_configurations=valid_configurations,
        )
        column_check_factory_classes: list[CheckFactory] = [
            MissingCheckFactory(),
            InvalidCheckFactory(),
            DuplicateCheckFactory(),
            UserDefinedMetricExpressionCheckFactory(),
            UserDefinedMetricQueryCheckFactory(),
            FreshnessCheckFactory(),
            SqlFunctionCheckFactory(),
        ]
        check: Check = self.__create_check(check_args, column_check_factory_classes)
        if check:
            self.checks.append(check)
        else:
            self.logs.error(
                message=f"Invalid column {check_args.check_type} check",
                location=check_args.location,
            )

    def __parse_missing_configurations(self, check_yaml: dict, column: str) -> MissingConfigurations | None:
        yaml_helper: YamlHelper = YamlHelper(self.logs)
        missing_values: list | None = yaml_helper.read_list_opt(check_yaml, "missing_values")
        missing_regex_sql: str | None = yaml_helper.read_string_opt(check_yaml, "missing_regex_sql")

        if all(v is None for v in [missing_values, missing_regex_sql]):
            return self.missing_value_configs_by_column.get(column)

        else:
            missing_configurations = MissingConfigurations(
                missing_values=missing_values, missing_regex_sql=missing_regex_sql
            )

            # If a missing config is specified, do a complete overwrite.
            # Overwriting the missing configs gives more control to the contract author over merging the missing configs.
            self.missing_value_configs_by_column[column] = missing_configurations

            return missing_configurations

    def __parse_valid_configurations(self, check_yaml: dict, column: str) -> ValidConfigurations | None:
        yaml_helper: YamlHelper = YamlHelper(self.logs)

        invalid_values: list | None = yaml_helper.read_list_opt(check_yaml, "invalid_values")
        invalid_format: str | None = yaml_helper.read_string_opt(check_yaml, "invalid_format")
        invalid_regex_sql: str | None = yaml_helper.read_string_opt(check_yaml, "invalid_regex_sql")

        valid_values: list | None = yaml_helper.read_list_opt(check_yaml, "valid_values")

        valid_format: str | None = yaml_helper.read_string_opt(check_yaml, "valid_format")
        valid_regex_sql: str | None = yaml_helper.read_string_opt(check_yaml, "valid_regex_sql")

        valid_min: Number | None = yaml_helper.read_number_opt(check_yaml, "valid_min")
        valid_max: Number | None = yaml_helper.read_number_opt(check_yaml, "valid_max")

        valid_length: int | None = yaml_helper.read_number_opt(check_yaml, "valid_length")
        valid_min_length: int | None = yaml_helper.read_number_opt(check_yaml, "valid_min_length")
        valid_max_length: int | None = yaml_helper.read_number_opt(check_yaml, "valid_max_length")

        valid_values_reference_data: ValidValuesReferenceData | None = None
        valid_values_reference_data_yaml_object: dict | None = yaml_helper.read_dict_opt(
            check_yaml, "valid_values_reference_data"
        )
        if valid_values_reference_data_yaml_object:
            ref_dataset = yaml_helper.read_string(valid_values_reference_data_yaml_object, "dataset")
            ref_column = yaml_helper.read_string(valid_values_reference_data_yaml_object, "column")
            valid_values_reference_data = ValidValuesReferenceData(dataset=ref_dataset, column=ref_column)

        if all(
            v is None
            for v in [
                invalid_values,
                invalid_format,
                invalid_regex_sql,
                valid_values,
                valid_format,
                valid_regex_sql,
                valid_min,
                valid_max,
                valid_length,
                valid_min_length,
                valid_max_length,
                valid_values_reference_data,
            ]
        ):
            return self.valid_value_configs_by_column.get(column)
        else:
            valid_configurations = ValidConfigurations(
                invalid_values=invalid_values,
                invalid_format=invalid_format,
                invalid_regex_sql=invalid_regex_sql,
                valid_values=valid_values,
                valid_format=valid_format,
                valid_regex_sql=valid_regex_sql,
                valid_min=valid_min,
                valid_max=valid_max,
                valid_length=valid_length,
                valid_min_length=valid_min_length,
                valid_max_length=valid_max_length,
                valid_values_reference_data=valid_values_reference_data,
            )

            # If a valid config is specified, do a complete overwrite.
            # Overwriting the valid configs gives more control to the contract author over merging the missing configs.
            self.valid_value_configs_by_column[column] = valid_configurations

            return valid_configurations

    def __parse_numeric_threshold(self, check_yaml: dict) -> Threshold | None:
        yaml_helper: YamlHelper = YamlHelper(self.logs)

        numeric_threshold: Threshold = Threshold(
            greater_than=yaml_helper.read_number_opt(check_yaml, "must_be_greater_than"),
            greater_than_or_equal=yaml_helper.read_number_opt(check_yaml, "must_be_greater_than_or_equal_to"),
            less_than=yaml_helper.read_number_opt(check_yaml, "must_be_less_than"),
            less_than_or_equal=yaml_helper.read_number_opt(check_yaml, "must_be_less_than_or_equal_to"),
            equal=yaml_helper.read_number_opt(check_yaml, "must_be"),
            not_equal=yaml_helper.read_number_opt(check_yaml, "must_not_be"),
            between=yaml_helper.read_range(check_yaml, "must_be_between"),
            not_between=yaml_helper.read_range(check_yaml, "must_be_not_between"),
        )

        for key in check_yaml:
            if key.startswith("must_") and key not in AbstractCheck.threshold_keys:
                self.logs.error(f"Invalid threshold '{key}'. Must be in '{AbstractCheck.threshold_keys}'.")

        return numeric_threshold

    def __create_check(self, check_args: CheckArgs, column_check_factory_classes: list[CheckFactory]) -> Check | None:
        for column_check_factory_class in column_check_factory_classes:
            check = column_check_factory_class.create_check(check_args)
            if check:
                return check


@dataclass
class ContractResult:
    """
    This is the immutable data structure containing all the results from a single contract verification.
    This includes any potential execution errors as well as the results of all the checks performed.
    """

    contract: Contract
    sodacl_yaml_str: str | None
    # self.logs combines all the logs of the contract verification with the logs of the Connection parsing,
    # connection usage, SodaCloud parsing and usage (if used) and contract parsing.
    # At the end of the verify method a SodaException is raised if there are any error logs or check failures.
    # See also adr/03_exceptions_vs_error_logs.md
    logs: Logs
    check_results: List[CheckResult]

    def __init__(
        self, data_source: ContractDataSource, contract: Contract, sodacl_yaml_str: str | None, logs: Logs, scan: Scan
    ):
        self.data_source_yaml_dict: dict = data_source.data_source_yaml_dict
        self.contract = contract
        self.sodacl_yaml_str = sodacl_yaml_str
        # See also adr/03_exceptions_vs_error_logs.md
        self.logs: Logs = Logs(logs)
        self.check_results: List[CheckResult] = []

        contract_checks_by_id: dict[str, Check] = {check.identity: check for check in contract.checks}

        schema_check: SchemaCheck | None = next((c for c in contract.checks if isinstance(c, SchemaCheck)), None)

        scan_metrics_by_id: dict[str, dict] = {
            scan_metric["identity"]: scan_metric for scan_metric in scan.scan_results.get("metrics", [])
        }

        scan_checks = scan.scan_results.get("checks")
        if isinstance(scan_checks, list):
            for scan_check in scan_checks:
                contract_check: Check | None = None
                if scan_check.get("name") == "Schema Check" and scan_check.get("type") == "generic":
                    contract_check = schema_check
                else:
                    source_identity = scan_check.get("source_identity")
                    if isinstance(source_identity, str):
                        contract_check = contract_checks_by_id[source_identity]

                assert contract_check is not None, "Contract scan check matching failed :("

                scan_check_metric_ids = scan_check.get("metrics")
                scan_check_metrics = [
                    scan_metrics_by_id.get(check_metric_id) for check_metric_id in scan_check_metric_ids
                ]
                scan_check_metrics_by_name = {
                    scan_check_metric.get("metricName"): scan_check_metric for scan_check_metric in scan_check_metrics
                }
                check_result = contract_check.create_check_result(
                    scan_check=scan_check, scan_check_metrics_by_name=scan_check_metrics_by_name, scan=scan
                )
                self.check_results.append(check_result)

    def failed(self) -> bool:
        """
        Returns true if there are checks that have failed.
        Ignores execution errors in the logs.
        """
        return any(check.outcome == CheckOutcome.FAIL for check in self.check_results)

    def passed(self) -> bool:
        """
        Returns true if there are no checks that have failed.
        Ignores execution errors in the logs.
        """
        return not self.failed()

    def __str__(self) -> str:
        error_texts_list: List[str] = [str(error) for error in self.logs.get_errors()]

        check_failure_message_list: list[str] = []

        check_failure_count: int = 0
        for check_result in self.check_results:
            if check_result.outcome == CheckOutcome.FAIL:
                result_str_lines = check_result.get_contract_result_str_lines()
                check_failure_message_list.extend(result_str_lines)
                check_failure_count += 1

        if not error_texts_list and not check_failure_message_list:
            return "All is good. No checks failed. No contract execution errors."

        errors_summary_text = f"{len(error_texts_list)} execution error"
        if len(error_texts_list) != 1:
            errors_summary_text = f"{errors_summary_text}s"

        checks_summary_text = f"{check_failure_count} check failure"
        if check_failure_count != 1:
            checks_summary_text = f"{checks_summary_text}s"

        parts = [f"{checks_summary_text} and {errors_summary_text}"]
        if error_texts_list:
            error_lines_text: str = indent("\n".join(error_texts_list), "  ")
            parts.append(f"Errors: \n{error_lines_text}")

        if check_failure_message_list:
            parts.append("\n".join(check_failure_message_list))

        return "\n".join(parts)



from __future__ import annotations

import json

from jsonschema.validators import Draft7Validator

from soda.contracts.impl.logs import Logs


class ValidatorLoader:

    @classmethod
    def load_json_schema_validator(cls) -> Draft7Validator:
        suffix = "/impl/json_schema_verifier.py"
        contracts_dir = __file__[: -len(suffix)]
        contract_schema_json_file_path = f"{contracts_dir}/soda_data_contract_json_schema_1_0_0.json"
        with open(contract_schema_json_file_path) as f:
            contract_schema_json_str = f.read()
            schema_dict = json.loads(contract_schema_json_str)
            return Draft7Validator(schema_dict)


class JsonSchemaVerifier:

    __validator = ValidatorLoader.load_json_schema_validator()

    def __init__(self, logs: Logs | None = None):
        # See also adr/03_exceptions_vs_error_logs.md
        self.logs: Logs = logs if logs else Logs()

    def verify(self, yaml_object: object) -> None:
        """
        Verifies that the YAML data structure matches the data contract schema.
        Swallows all errors and exceptions and appends them to self.logs.
        """
        for error in self.__validator.iter_errors(instance=yaml_object):
            error_path_text = "contract document level" if len(error.path) == 0 else error.json_path
            self.logs.error(f"JSON schema error: {error.message} ({error_path_text})")
